package lsedit;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.undo.UndoableEdit;

public class HistoryBox extends JComponent implements ToolBarEventHandler, MouseListener, MouseMotionListener
{
	public class ShowCompoundEdit extends JDialog { 

		public class ShowCompoundEditPanel extends JComponent {
			
			Vector m_v;

			ShowCompoundEditPanel(Vector v)
			{
				m_v = v;
			}

			public void paintComponent(Graphics g)
			{
				int height = HistoryBox.HEIGHT;
				Vector v   = m_v;

				setBackground(Color.lightGray);

				if (v != null) {
					Enumeration		en;
					UndoableEdit	undoableEdit;

					g.setFont(m_textFont);
					g.setColor(Color.BLACK);

					for (en = v.elements(); en.hasMoreElements(); ) {
						undoableEdit = (UndoableEdit) en.nextElement();
						if (undoableEdit instanceof MyPaintableUndoableEdit) {
							((MyPaintableUndoableEdit) undoableEdit).paintComponent(g, MARGIN, height);
						} else {
							g.drawString(undoableEdit.getPresentationName(), MARGIN, height+BASELINE);
						}
						height      += HistoryBox.HEIGHT;
			}	}	}
		}

		MyCompoundEdit		m_compoundEdit;

		public ShowCompoundEdit(LandscapeEditorCore ls, HistoryBox historyBox, MyCompoundEdit compoundEdit)
		{
			super(ls.getFrame(), compoundEdit.getPresentationName(), true);

			int				x, y, width, height, w;
			Enumeration		en;
			UndoableEdit	undoableEdit;
			Vector			v;
			Dimension		preferredSize = new Dimension();
			Container		contentPane;
			ShowCompoundEditPanel panel;

			x      = historyBox.m_tabbedPane.getX() + 20;
			y      = historyBox.m_tabbedPane.getY() + 20;
			setLocation(x,y);

			width  = historyBox.getWidth();
			height = historyBox.getHeight();
			m_compoundEdit = compoundEdit;
			v              = compoundEdit.getEdits();

			preferredSize.height = (v.size() * HistoryBox.HEIGHT) + HistoryBox.GAP + 20;
			preferredSize.width  = 0;
			for (en = v.elements(); en.hasMoreElements(); ) {
				undoableEdit = (UndoableEdit) en.nextElement();

				if (undoableEdit instanceof MyPaintableUndoableEdit) {
					w   = ((MyPaintableUndoableEdit) undoableEdit).getPreferredWidth();
				} else {
					FontMetrics	fm;

					fm   = getFontMetrics(HistoryBox.m_textFont);
					w    = fm.stringWidth(undoableEdit.getPresentationName());
				}
				if (w > preferredSize.width) {
					preferredSize.width = w;
			}	}
			preferredSize.width  += HistoryBox.GAP;
			if (preferredSize.width < width / 2) {
				preferredSize.width = width / 2;
			}
			contentPane = getContentPane();
			contentPane.setLayout(new BorderLayout());
			contentPane.setSize(preferredSize);
			panel       = new ShowCompoundEditPanel(v);
						panel.setPreferredSize(preferredSize);
			panel.setSize(preferredSize);
			contentPane.add(panel, BorderLayout.CENTER);

			pack();
			show();
		}
	}

	public class ShowMaxHistory extends JDialog implements ActionListener {

		protected LandscapeEditorCore	m_ls;
		protected Diagram				m_diagram;
		protected HistoryBox			m_historyBox;
		protected JTextField			m_max;
		protected int					m_limit;
		protected JButton				m_ok;
		protected JButton				m_cancel;
		
		public ShowMaxHistory(LandscapeEditorCore ls, HistoryBox historyBox)
		{
			super(ls.getFrame(), "Maximum History", true);

			JFrame			frame;
			Container		contentPane;
			JPanel			grid;
			JPanel			panel;
			JLabel			label;
			int				x, y;
			
			m_ls          = ls;
			m_diagram     = ls.getDiagram();

			frame = ls.getFrame();

			x      = historyBox.m_tabbedPane.getX() + 20;
			y      = historyBox.m_tabbedPane.getY() + 20;
			setLocation(x, y);
			setForeground(ColorCache.get(0,0,0));
			setBackground(ColorCache.get(192,192,192));
			setFont(FontCache.get("Dialog",0,12));

			contentPane = getContentPane();
			
			grid           = new JPanel();
			grid.setLayout(new GridLayout(0,1));

			panel = new JPanel();
			panel.setLayout(new FlowLayout());
			label      = new JLabel("Maximum: ");
			panel.add(label);
			m_limit = m_diagram.getLimit();
			m_max = new JTextField("" + m_limit, 10);
			panel.add(m_max);
			grid.add(panel);
			
			contentPane.add(grid, BorderLayout.CENTER);
			 
			panel = new JPanel();
			panel.setLayout(new FlowLayout());

			m_ok = new JButton("Ok");
			panel.add(m_ok);
			m_ok.addActionListener(this);
			m_cancel = new JButton("Cancel");
			panel.add(m_cancel);
			m_cancel.addActionListener(this);

			contentPane.add(panel, BorderLayout.SOUTH);

			// Resize the window to the preferred size of its components

			this.pack();
			show();
		}

		// ActionListener interface

		public void actionPerformed(ActionEvent ev)
		{
			Object	source;

			// Pop down the window when the button is clicked.
			// System.out.println("event: " + ev);

			source = ev.getSource();

			if (source == m_ok) {
				String	max;

				max = m_max.getText();
				max = max.trim();
				if (max.length() > 0) {
					try {
						int		limit;

						limit = Integer.parseInt(max);
						if (limit >= 0) {
							m_diagram.setLimit(limit);
							m_historyBox.fillHistoryBox();
						}
					} catch (Throwable exception) {
				}	}
				this.setVisible(false);
				return;
			}

			if (source == m_cancel) {
				this.setVisible(false);
				return;
			}
		}
	}

	public class ShowReallyDelete extends JDialog implements ActionListener {

		protected LandscapeEditorCore	m_ls;
		protected HistoryBox			m_historyBox;
		protected JButton				m_ok;
		protected JButton				m_cancel;
		
		public ShowReallyDelete(LandscapeEditorCore ls, HistoryBox historyBox)
		{
			super(ls.getFrame(), "Delete History", true);

			JFrame			frame;
			Container		contentPane;
			JPanel			grid;
			JPanel			panel;
			JLabel			label;
			int				x, y;
			
			m_ls          = ls;
			m_historyBox  = historyBox;

			frame = ls.getFrame();

			x      = historyBox.m_tabbedPane.getX() + 20;
			y      = historyBox.m_tabbedPane.getY() + 20;
			setLocation(x, y);
			setForeground(ColorCache.get(0,0,0));
			setBackground(ColorCache.get(192,192,192));
			setFont(FontCache.get("Dialog",Font.PLAIN,12));

			contentPane = getContentPane();
			
			grid           = new JPanel();
			grid.setLayout(new GridLayout(0,1));

			label          = new JLabel("Confirm deletion?");
			grid.add(label);
			
			contentPane.add(grid, BorderLayout.CENTER);
			 
			panel = new JPanel();
			panel.setLayout(new FlowLayout());

			m_ok = new JButton("Ok");
			panel.add(m_ok);
			m_ok.addActionListener(this);
			m_cancel = new JButton("Cancel");
			panel.add(m_cancel);
			m_cancel.addActionListener(this);

			contentPane.add(panel, BorderLayout.SOUTH);

			// Resize the window to the preferred size of its components

			this.pack();
			show();
		}

		// ActionListener interface

		public void actionPerformed(ActionEvent ev)
		{
			Object	source;

			// Pop down the window when the button is clicked.
			// System.out.println("event: " + ev);

			source = ev.getSource();

			if (source == m_ok || source == m_cancel) {
				if (source == m_ok) {
					m_ls.discardAllEdits();
					m_historyBox.fillHistoryBox();
				}
				this.setVisible(false);
				return;
		}	}
	}

	protected static final int MARGIN    = 5;
	protected static final int HEIGHT    = 16;
	protected static final int BASELINE  = 12;
	protected static final int GAP       = 5; 

	protected static final int TY_CLEAR	 = 0;
	protected static final int TY_RAISED = 1;
	protected static final int TY_SUNK	 = 2;

	protected LandscapeEditorCore   m_ls;
	protected JTabbedPane			m_tabbedPane;
	protected JScrollPane			m_scrollPane;

	public static Dimension         m_preferredSize    = new Dimension(0,0);

	public    static final String m_helpStr	 = "This box shows the history of updates made in the current diagram.";

	protected int					m_style;
	protected int					m_over;

	public final static String DEFAULT_HISTORY_FONT_NAME  = FontCache.DEFAULT_FONT_NAME;
	public final static int    DEFAULT_HISTORY_FONT_STYLE = Font.PLAIN;
	public final static int    DEFAULT_HISTORY_FONT_SIZE  = 11;

	protected static     Font		m_textFont	 = FontCache.get(DEFAULT_HISTORY_FONT_NAME, DEFAULT_HISTORY_FONT_STYLE, DEFAULT_HISTORY_FONT_SIZE);
	// --------------
	// Public methods 
	// --------------

	public HistoryBox(LandscapeEditorCore ls, JTabbedPane tabbedPane) 
	{
		m_ls          = ls;
		m_tabbedPane  = tabbedPane;

		setBackground(Diagram.boxColour);
//		setToolTipText(m_helpStr);

		m_scrollPane = new JScrollPane();
		m_scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED); 
		m_scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED); 
		m_scrollPane.setViewportView(this);
		tabbedPane.addTab("Undo", null, m_scrollPane, m_helpStr);
		addMouseListener(this);
		addMouseMotionListener(this);
	}

	public static Font getTextFont()
	{
		return m_textFont;
	}

	public static void setTextFont(Font font)
	{
		m_textFont = font;
	}

	public void textFontChanged(Font font)
	{
		fillHistoryBox();
	}

	public void setNewPreferredSize()
	{
//		System.out.println("HistoryBox.setNewPreferredSize " + m_preferredSize);
		setPreferredSize(m_preferredSize);
		setSize(m_preferredSize);
		m_scrollPane.revalidate();
	}

/*
	public Dimension getPreferredSize()
	{
		Dimension d = super.getPreferredSize();

		System.out.println("HistoryBox.getPreferredSize " + d);
		return d;
	}

	public Dimension getSize()
	{
		Dimension d = super.getSize();

		System.out.println("HistoryBox.getSize " + d);
		return d;
	}
*/

	public LandscapeEditorCore getLegendLs()
	{
		return(m_ls);
	}

	public void fillHistoryBox()
	{
		repaint();
	}

	public void activate() 
	{
		fillHistoryBox();
		m_tabbedPane.setSelectedComponent(m_scrollPane);
	}

	public boolean isActive() 
	{
		Component active;

		if (!isVisible()) {
			return(false);
		}
		active = m_tabbedPane.getSelectedComponent();
		return(active != null && m_scrollPane == active);
	}

	public Dimension getMaximumSize()
	{
		return(getPreferredSize());
	}

	public void paintComponent(Graphics g)
	{
		int height = HEIGHT;
		Vector v   = m_ls.getEdits();

		if (v != null) {
			Enumeration		en;
			UndoableEdit	undoableEdit;
			UndoableEdit	editToBeRedone = m_ls.getEditToBeRedone();
			String			name;
			int				at     = m_over;

			g.setFont(m_textFont);
			g.setColor(Color.BLACK);

			for (en = v.elements(); en.hasMoreElements(); --at) {
				undoableEdit = (UndoableEdit) en.nextElement();
				if (undoableEdit == editToBeRedone) {
					g.setColor(Color.RED);
				}
				if (at == 0) {
					if (m_style != MapBox.TY_CLEAR) {
						Color color;

						color = g.getColor();
						g.setColor(Color.BLACK);
						g.draw3DRect(0, height, m_preferredSize.width-1, HEIGHT-2, m_style == TY_RAISED);
						g.setColor(color);
				}	} 
				if (undoableEdit instanceof MyPaintableUndoableEdit) {
					((MyPaintableUndoableEdit) undoableEdit).paintComponent(g, MARGIN, height);
				} else if (undoableEdit instanceof MyCompoundEdit) {
					Color color;

					color = g.getColor();
					g.setColor(Color.WHITE);
					g.fill3DRect(0, height, m_preferredSize.width-1, HEIGHT-2, (at == 0 && m_style == TY_RAISED));
					g.setColor(color);
					g.drawString(undoableEdit.getPresentationName(), MARGIN, height+BASELINE);
				} else {
					g.drawString(undoableEdit.getPresentationName(), MARGIN, height+BASELINE);
				}
				height      += HEIGHT;
	}	}	}


	protected void mouseAt(int y)
	{
		int	over = (y / HEIGHT) - 1;
		int	max;
		int style;

		if (over < 0) {
			style = HistoryBox.TY_CLEAR;
			over  = -1;
		} else if (over >= (max = m_ls.countEdits())) {
			style = HistoryBox.TY_CLEAR;
			over  = max;
		} else {
			style = HistoryBox.TY_RAISED;
		}
		if (style != m_style || over != m_over) {
			m_style = style;
			m_over  = over;
			repaint();
	}	}

	// MouseListener interface

	public void mouseClicked(MouseEvent ev)
	{
	}

	public void mouseEntered(MouseEvent ev)
	{
		mouseAt(ev.getY());
	}

	public void mouseExited(MouseEvent ev)
	{
		m_style = MapBox.TY_CLEAR;
		repaint();
	}

	public void mousePressed(MouseEvent ev)
	{
		int over;

		m_over  = over = (ev.getY() / HEIGHT) - 1;
		if (ev.isMetaDown()) {
			JPopupMenu	m =  new JPopupMenu("History options");
			Vector		v = m_ls.getEdits();
			Diagram		diagram;
			
			diagram = m_ls.getDiagram();
			if (diagram != null) {
				if (v != null) {
					int			max;

					max = v.size();
					if (max > 0) {

						if (over >= 0 && over < max) {
							UndoableEdit	undoableEdit = (UndoableEdit) v.elementAt(over);

							if (undoableEdit instanceof MyCompoundEdit) {
								MyMenuItem	m1;
						
								m1 = new MyMenuItem(m, "Show details", this, -1, Do.DESCEND);
								m1.setObject(undoableEdit);
							}
						} 
						new MyMenuItem(m, "Dispose", this, -1, Do.DELETE);
				}	}
				new MyMenuItem(m, "Limit", this, -1, 'l');
				if (diagram.useCompoundEdit()) {
					new MyMenuItem(m, "No Compound Edits", this, -1, (int) 'C');
				} else {
					new MyMenuItem(m, "Use Compound Edits", this, -1, (int) 'C');
				}
				FontCache.setMenuTreeFont(m); 
				m.show(this, ev.getX(), ev.getY());
			}
			return;
		}

		m_style = MapBox.TY_SUNK;
		if (over >= 0) {
			Vector v   = m_ls.getEdits();

			if (v != null) {
				Enumeration		en;
				UndoableEdit	undoableEdit;
				UndoableEdit	editToBeRedone = m_ls.getEditToBeRedone();
				String			name;
				int				at     = m_over;
				boolean			redo   = false;

				for (en = v.elements(); en.hasMoreElements(); --at) {
					undoableEdit = (UndoableEdit) en.nextElement();
					if (undoableEdit == editToBeRedone) {
						redo = true;
					}
					if (at == 0) {
						m_ls.massChange(undoableEdit, redo);
						break;
		}	}	}	}
		repaint();
	}

	public void mouseReleased(MouseEvent ev)
	{

		m_style = MapBox.TY_CLEAR;
		repaint();
	}

	// MouseMotionListener interface

	public void mouseDragged(MouseEvent ev)
	{
		mouseAt(ev.getY());

	}

	public void mouseMoved(MouseEvent ev)
	{
		mouseAt(ev.getY());
	}

	// ToolBarEventHandler

	 public void processKeyEvent(int key, int modifiers, Object object)
	 {
		switch (key) {
		case Do.DESCEND:
			new ShowCompoundEdit(m_ls, this, (MyCompoundEdit) object);
			break;
		case Do.DELETE:
			new ShowReallyDelete(m_ls, this);
			break;
		case 'C':
		case 'c':
		{
			Diagram diagram = m_ls.getDiagram();
			if (diagram != null) {
				boolean state = diagram.useCompoundEdit();
				diagram.setUseCompoundEdit(!state);
			}
			break;
		}
		case 'L':
		case 'l':
			new ShowMaxHistory(m_ls, this);
			break;

	 }	}

	 public void showInfo(String msg)
	 {
	 }		

}



